/* 
 * HPCU Bi-Weekly Challenge: Hybrid Parallel Rumor Mill
 * Author: Aaron Weeden, Shodor, 2014
 * Last Modified: November 19, 2014, by Aaron Weeden
 */

/* import libraries */
#include <omp.h>
#include <math.h>
#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>

/* create macro for exiting if MPI errors occur */
#define tryMPI(ret, str, rank) \
    if (ret != MPI_SUCCESS) { \
        fprintf(stderr, "Rank %d returned error for %s, exiting!\n", \
                rank, str); \
        exit(EXIT_FAILURE); \
    }

/* declare constants */
const int MPI_TAG = 0;
const double INITIAL_GULLIBLES = 999;
const double INITIAL_RUMOR_MONGERERS = 1;
const double INITIAL_LOYALISTS = 0;
const double START_TIME = 0.0;
const double END_TIME = 50.0;
const double TIME_STEP = 0.015625;

/* start program */
int main(int argc, char **argv) {

    /* declare variables */
    /* MPI and OpenMPI */
    int mpi_rank;
    int mpi_size;
    int thread_num;
    int num_threads;
    /* simulation parameters */
    double spreading_probability;
    double rationality_rate;
    /* quantities */
    double gullibles;
    double rumor_mongerers;
    double loyalists;
    /* rates */
    double mongerization;
    double coming_to_senses;
    /* final amounts */
    double *final_gullibles;
    double *final_rumor_mongerers;
    double *final_loyalists;
    /* loop control */
    int i;
    int j;
    double current_time_step;

    /* set up mpi */
    tryMPI(
            MPI_Init(&argc, &argv),
            "MPI_Init",
            0
          );
    tryMPI(
            MPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank),
            "MPI_Comm_rank",
            0
          );
    tryMPI(
            MPI_Comm_size(MPI_COMM_WORLD, &mpi_size),
            "MPI_Comm_size",
            mpi_rank
          );

    /* get number of threads */
#pragma omp parallel
    {
#pragma omp single
        {
            num_threads = omp_get_num_threads();
        }
    }

    /* allocate memory for arrays (enough space for each thread) */
    if ((final_gullibles
                = malloc(num_threads * sizeof(double))) == NULL) {
        fprintf(stderr,
                "Rank %d returned error for final_gullibles malloc, exiting!\n",
                mpi_rank);
        exit(EXIT_FAILURE);
    }
    if ((final_rumor_mongerers
                = malloc(num_threads * sizeof(double))) == NULL) {
        fprintf(stderr,
                "Rank %d returned error for final_rumor_mongerers malloc, exiting!\n",
                mpi_rank);
        exit(EXIT_FAILURE);
    }
    if ((final_loyalists = malloc(num_threads * sizeof(double))) == NULL) {
        fprintf(stderr,
                "Rank %d returned error for final_loyalists malloc, exiting!\n",
                mpi_rank);
        exit(EXIT_FAILURE);
    }

    /* initialize parameters that depend on MPI rank */
    if (mpi_size == 1)
        spreading_probability = 0.0;
    else
        spreading_probability = 0.005 * mpi_rank / (mpi_size - 1);

    /* spawn OpenMP threads */
#pragma omp parallel private( \
        rationality_rate, \
        gullibles, \
        rumor_mongerers, \
        loyalists, \
        mongerization, \
        coming_to_senses, \
        thread_num, \
        current_time_step \
        )
    {
        /* initialize parameters that depend on OpenMP thread number */
        thread_num = omp_get_thread_num();
        if (num_threads == 1)
            rationality_rate = 0.0;
        else
            rationality_rate = (double)thread_num / (num_threads - 1);

        /* initialize quantities */
        gullibles = INITIAL_GULLIBLES;
        rumor_mongerers = INITIAL_RUMOR_MONGERERS;
        loyalists = INITIAL_LOYALISTS;

        /* initialize rates */
        mongerization = coming_to_senses = 0.0;

        /* run the simulation */
        for (current_time_step = START_TIME; current_time_step <= END_TIME;
                current_time_step += TIME_STEP) {

            /* update quantities */
            gullibles -= TIME_STEP * mongerization;
            rumor_mongerers += TIME_STEP * (mongerization - coming_to_senses);
            loyalists += TIME_STEP * coming_to_senses;

            /* update rates */
            mongerization = spreading_probability * gullibles * rumor_mongerers;
            coming_to_senses = rationality_rate * rumor_mongerers;
        }

        /* store the final quantities in arrays */
        final_gullibles[thread_num] = gullibles;
        final_rumor_mongerers[thread_num] = rumor_mongerers;
        final_loyalists[thread_num] = loyalists;

        /* done spawning threads */
    }

    /* send arrays to rank 0 */
    tryMPI(
            MPI_Send(final_gullibles, num_threads, MPI_DOUBLE, 0, MPI_TAG,
                MPI_COMM_WORLD),
            "MPI_Send",
            mpi_rank
          );
    tryMPI(
            MPI_Send(final_rumor_mongerers, num_threads, MPI_DOUBLE, 0, MPI_TAG,
                MPI_COMM_WORLD),
            "MPI_Send",
            mpi_rank
          );
    tryMPI(
            MPI_Send(final_loyalists, num_threads, MPI_DOUBLE, 0, MPI_TAG,
                MPI_COMM_WORLD),
            "MPI_Send",
            mpi_rank
          );

    /* rank 0 only */
    if (mpi_rank == 0) {

        /* receive arrays from each process */
        for (i = 0; i < mpi_size; i++) {
            tryMPI(
                    MPI_Recv(final_gullibles, num_threads, MPI_DOUBLE, i,
                        MPI_TAG, MPI_COMM_WORLD, MPI_STATUS_IGNORE),
                    "MPI_Recv",
                    mpi_rank
                  );
            tryMPI(
                    MPI_Recv(final_rumor_mongerers, num_threads, MPI_DOUBLE, i,
                        MPI_TAG, MPI_COMM_WORLD, MPI_STATUS_IGNORE),
                    "MPI_Recv",
                    mpi_rank
                  );
            tryMPI(
                    MPI_Recv(final_loyalists, num_threads, MPI_DOUBLE, i,
                        MPI_TAG, MPI_COMM_WORLD, MPI_STATUS_IGNORE),
                    "MPI_Recv",
                    mpi_rank
                  );

            /* print results */
            for (j = 0; j < num_threads; j++) {
                printf("%f %f %f %f %f\n",
                        0.005 * i / (mpi_size - 1),
                        (double)j / (num_threads - 1),
                        final_gullibles[j],
                        final_rumor_mongerers[j],
                        final_loyalists[j]);
            }
        }
    }

    /* free memory for arrays */
    free(final_loyalists);
    free(final_rumor_mongerers);
    free(final_gullibles);

    /* tear down mpi */
    tryMPI(
            MPI_Finalize(),
            "MPI_Finalize",
            mpi_rank
          );

    /* quit program */
    return 0;
}

